{ Description:

  This unit provides 'Large' collection objects that go beyond the
  16,000 item limit imposed by OWL collections.

  These 'large' collections allow up to MaxLongInt items per collection.

  The unit includes a 'TLCollection', a 'TLSortedCollection' and a
  'TLStrCollection' all of which are functionally identical to their
  OWL equivalents.

  I've added some validation routines to the methods that process
  collection element numbers, so that its no longer possible to
  crash an app by posting an erroneous element number.

  I'm indebted to Peter Sawatzki [10031,3002] for his help on the
  low level code for moving data blocks within global memory blocks.
  Most of the low level code has been taken from Peter's excellent
  BpTools unit available from this forum library.

  Any bugs, questions, suggestions should be directed to:

      Ian Hayes,
      Soft Systems Ltd,
      4 Berghem Mews,
      Blythe Road,
      London W14, UK
      Compuserve : 100010,1415
      CIX        : ICFM
      Voice      : UK 071-602-3010

}

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

UNIT LargeCol;   { Large collection - up to MaxLongInt items }

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

INTERFACE

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

USES

   Objects,
   Strings,
   WinTypes,
   WinProcs;

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

CONST

  MaxLCollectionSize = MaxLongInt div SizeOf(Pointer);

  { Large Collection Error codes }

  lce_OK         =  0;   { status OK }
  lce_IdxLtZero  = -1;   { Index parameter value less than zero }
  lce_IdxGtCount = -2;   { Index parameter value greater than count }
  lce_MemHFail   = -3;   { GlobalAlloc memory request failed }
  lce_Empty      = -4;   { trying to put or del an item in
                           an empty collection }
  lce_AtPutIdx   = -5;   { trying to AtPut beyond Count }

TYPE

   { --- Large Collection --- }

   PLCollection = ^TLCollection;
   TLCollection = OBJECT(TObject)
      MemH      : THandle;
      Address   : Pointer;
      Count     : LONGINT;
      Limit     : LONGINT;
      Delta     : LONGINT;
      Status    : LONGINT;
      ErrorInfo : LONGINT;
      CONSTRUCTOR Init(ALimit, ADelta: LONGINT);
      CONSTRUCTOR Load(VAR S: TStream);
      DESTRUCTOR Done; VIRTUAL;
      FUNCTION At(Index: LONGINT): Pointer;
      PROCEDURE AtDelete(Index: LONGINT);
      PROCEDURE AtFree(Index: LONGINT);
      PROCEDURE AtInsert(Index: LONGINT; Item: Pointer);
      PROCEDURE AtPut(Index: LONGINT; Item: Pointer);
      PROCEDURE Delete(Item: Pointer);
      PROCEDURE DeleteAll;
      PROCEDURE Error(Code,Info: LONGINT); VIRTUAL;
      FUNCTION FirstThat(Test: Pointer): Pointer;
      PROCEDURE ForEach(Action: Pointer);
      PROCEDURE Free(Item: Pointer);
      PROCEDURE FreeAll;
      PROCEDURE FreeItem(Item: Pointer); VIRTUAL;
      FUNCTION GetItem(VAR S: TStream): Pointer; VIRTUAL;
      FUNCTION IndexOf(Item: Pointer): LONGINT; VIRTUAL;
      PROCEDURE Insert(Item: Pointer); VIRTUAL;
      FUNCTION LastThat(Test: Pointer): Pointer;
      FUNCTION MemPtr(Idx: LONGINT) : POINTER;
      PROCEDURE Pack;
      PROCEDURE PutItem(VAR S: TStream; Item: Pointer); VIRTUAL;
      FUNCTION SetLimit(ALimit: LONGINT) : BOOLEAN; VIRTUAL;
      PROCEDURE Store(VAR S: TStream);
      PRIVATE
      FUNCTION ValidIdx(Idx: LONGINT) : BOOLEAN;
   END;

   { --- Large Sorted Collection --- }

   PLSortedCollection = ^TLSortedCollection;
   TLSortedCollection = OBJECT(TLCollection)
      Duplicates: Boolean;
      CONSTRUCTOR Init(ALimit, ADelta: LONGINT);
      CONSTRUCTOR Load(VAR S: TStream);
      FUNCTION Compare(Key1, Key2: Pointer): INTEGER; VIRTUAL;
      FUNCTION IndexOf(Item: Pointer): LONGINT; VIRTUAL;
      PROCEDURE Insert(Item: Pointer); VIRTUAL;
      FUNCTION KeyOf(Item: Pointer): Pointer; VIRTUAL;
      FUNCTION Search(Key: Pointer; VAR Index: LONGINT): Boolean; VIRTUAL;
      PROCEDURE Store(VAR S: TStream);
   END;

   { --- Large String Collection --- }

   PLStrCollection = ^TLStrCollection;
   TLStrCollection = OBJECT(TLSortedCollection)
      FUNCTION Compare(Key1, Key2: Pointer): Integer; VIRTUAL;
      PROCEDURE FreeItem(Item: Pointer); VIRTUAL;
      FUNCTION GetItem(var S: TStream): Pointer; VIRTUAL;
      PROCEDURE PutItem(var S: TStream; Item: Pointer); VIRTUAL;
   END;

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

IMPLEMENTATION

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

Const

   AHi: Word = 8;
   AHs: Byte = 3;

{ :::::::::::::::::: Unit level private procs/funcs :::::::::::::::::::: }

Procedure AHIncr;  FAR; External 'KERNEL' Index 114; {magic function}
Procedure AHShift; FAR; External 'KERNEL' Index 113; {dito}

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

Function MaxFit (Size: Word; anOffset: Word): Word;
{-return maximum number of bytes that fit in a Segment}
Inline(
  $5B/                   {  Pop Bx       ; anOffset}
  $58/                   {  Pop Ax       ; Size}
  $01/$C3/               {  Add Bx,Ax}
  $73/$02/               {  Jnc @@1}
  $29/$D8);              {  Sub Ax,Bx    ; $1.0000-anOffset}
                         {@@1:           ; Ax= Min(Size,$1.0000-anOffset)}

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

Function MaxChunk (Size: LongInt): Word;
{-return maximum number of bytes that can be transferred
  in one block using conventional functions}
Inline(
  $5B/                   {  Pop Bx       ; Word(Size)}
  $5A/                   {  Pop Dx       ; Word(Size+2)}
  $B8/$FF/$FF/           {  Mov Ax,$FFFF}
  $09/$D2/               {  Or Dx,Dx     ; Dx=0 ?}
  $75/$02/               {  Jne @@1}
  $89/$D8);              {  Mov Ax,Bx}
                         {@@1:           ; Ax=Min($FFFF,Size)}

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

Function LIncPtr (aPtr: Pointer; anOffset: LongInt): Pointer;
Assembler;
Asm
  Mov Dx,Word(anOffset+2)
  Mov Ax,Word(anOffset)
  Mov Cx,OFFSET AHShift
  Shl Dx,Cl
  Add Dx,Word(aPtr+2)
  Add Ax,Word(aPtr)
  Jnc @@1
  Add Dx,Offset AHincr
@@1:
End;

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

Function LDecPtr (aPtr: Pointer; anOffset: LongInt): Pointer;
Assembler;
Asm
  Mov Bx,Word(anOffset+2)
  Mov Cx,Offset AHShift
  Shl Bx,Cl
  Mov Dx,Word(aPtr+2)
  Mov Ax,Word(aPtr)
  Sub Dx,Bx
  Sub Ax,Word(anOffset)
  Jnc @@1
  Sub Dx,Offset AHincr
@@1:
End;

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

Function IncPtrMac (aPtr: Pointer; anOffset: Word): Pointer;
Inline(
  $5B/                   {  Pop Bx    ; anOffset}
  $58/                   {  Pop Ax    ; Word(aPtr)}
  $5A/                   {  Pop Dx    ; Word(aPtr+2)}
  $01/$D8/               {  Add Ax,Bx}
  $73/$04/               {  Jnc @@1}
  $03/$16/>AHI);         {  Add Dx,[>AHi]}
                         {@@1:}

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

Function DecPtrMac (aPtr: Pointer; anOffset: Word): Pointer;
Inline(
  $5B/                   {  Pop Bx    ; anOffset}
  $58/                   {  Pop Ax    ; Word(aPtr)}
  $5A/                   {  Pop Dx    ; Word(aPtr+2)}
  $29/$D8/               {  Sub Ax,Bx}
  $73/$04/               {  Jnc @@1}
  $2B/$16/>AHI);         {  Sub Dx,[>AHi]}
                         {@@1:}

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

Function PtrDiff (Ptr1,Ptr2: Pointer): LongInt; Assembler;
Asm
  Mov Dx,Word(Ptr1+2)
  Mov Bx,Word(Ptr2+2)
  Mov Cx,Offset AHshift
  Shr Dx,Cl
  Shr Bx,Cl
  Mov Ax,Word(Ptr1)
  Sub Ax,Word(Ptr2)
  Sbb Dx,Bx
  Jnc @@1
  Neg Ax
  Adc Dx,0
  Neg Dx
@@1:
END;

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE hMove(srcPtr, dstPtr: Pointer; Size: LongInt);
VAR
   Count: Word;
   S : LONGINT;

   { +++++++++++++++++++++++++++++++++++++++++++++++++++ }

   Function Min (a,b: Word): Word;
   BEGIN
      IF a > b THEN
         Min:= b
      ELSE
         Min:= a
   END;

   { +++++++++++++++++++++++++++++++++++++++++++++++++++ }

BEGIN
   S := Size;
   IF LongInt(dstPtr) > LongInt(srcPtr) THEN
   BEGIN {Shift up}
     srcPtr:= LIncPtr(srcPtr,Size-1);
     dstPtr:= LIncPtr(dstPtr,Size-1);
     WHILE Size > 0 DO
     BEGIN
       Count:= Min(Min(MaxChunk(Size),MaxChunk(LONGINT(Word(srcPtr))+1)),
                                      MaxChunk(LONGINT(Word(dstPtr))+1));
       srcPtr:= DecPtrMac(srcPtr, Count-1);
       dstPtr:= DecPtrMac(dstPtr, Count-1);
       Move(srcPtr^, dstPtr^, Count);
       srcPtr:= DecPtrMac(srcPtr, 1);
       dstPtr:= DecPtrMac(dstPtr, 1);
       Dec(Size, Count);
     END
   END
   ELSE {shift down}
   BEGIN
      WHILE Size > 0 DO
      BEGIN
         Count:= MaxFit(
                 MaxFit(MaxChunk(Size),Word(srcPtr))
                                      ,Word(dstPtr));
         Move(srcPtr^, dstPtr^, Count);
         srcPtr:= IncPtrMac(srcPtr,Count);
         dstPtr:= IncPtrMac(dstPtr,Count);
         Dec(Size,Count)
      END
   END
END;

{ ::::::::::::::::: TLCollection methods :::::::::::::::::::::::::::::::: }

CONSTRUCTOR TLCollection.Init(ALimit, ADelta: LONGINT);
BEGIN
   TObject.Init;
   MemH := 0;
   Address := NIL;
   Count := 0;
   Limit := 0;
   Status := lce_OK;
   ErrorInfo := 0;
   Delta := ADelta;
   IF NOT SetLimit(ALimit) THEN
      Fail;
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

CONSTRUCTOR TLCollection.Load(VAR S: TStream);
VAR
   I: LONGINT;
BEGIN
   { read/set object's data variables }
   S.Read(Count,SizeOf(Count));
   S.Read(Limit,SizeOf(Limit));
   S.Read(Delta,SizeOf(Delta));
   MemH := 0;
   Address := NIL;
   Status := lce_OK;
   ErrorInfo := 0;
   { get memory for collection }
   IF NOT SetLimit(Limit) THEN
      Fail
   ELSE
   BEGIN
      FOR I := 0 TO (Count-1) DO
         AtPut(I,GetItem(S));
   END
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

DESTRUCTOR TLCollection.Done;
BEGIN
   FreeAll;
   SetLimit(0);
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

{ Description: Returns a pointer to the 'Index' item in the large
               collection. Index is counted from base zero. }

FUNCTION TLCollection.At(Index: LONGINT): Pointer;
VAR
   SPtr,TPtr : POINTER;
BEGIN
   IF ValidIdx(Index) THEN
   BEGIN
      SPtr := MemPtr(Index);
      MOVE(SPtr^,TPtr,SIZEOF(POINTER));
      At := TPtr;
   END
   ELSE
      At := NIL;
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.AtDelete(Index: LONGINT);
VAR
   SPtr,TPtr : POINTER;
BEGIN
   { is it a valid index number? }
   IF Count = 0 THEN
      Error(lce_Empty,Index)
   ELSE
   BEGIN
      IF ValidIdx(Index) THEN
      BEGIN
         { IF its not the last item in the list THEN }
         IF Index <> (Count - 1) THEN
         BEGIN
            { shift all higher items back one place }
            TPtr := MemPtr(Index);
            SPtr := MemPtr(Index+1);
            hMove(SPtr,TPtr,(Count-1-Index)*SIZEOF(Pointer));
         END;
         { set the last element to zero }
         SPtr := MemPtr(Count-1);
         FILLCHAR(SPtr^,SIZEOF(POINTER),0);
         { dec count }
         DEC(Count);
      END
   END
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.AtFree(Index: LONGINT);
VAR
   Item: POINTER;
BEGIN
   Item := At(Index);
   AtDelete(Index);
   FreeItem(Item);
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.AtInsert(Index: LONGINT; Item: Pointer);
VAR
   S : BOOLEAN;
   SPtr,TPtr : POINTER;
BEGIN
   IF ValidIdx(Index) THEN
   BEGIN
      { is the collection full?
        IF so THEN increase size by delta }
      IF Count = Limit THEN
         S := SetLimit(Limit+Delta)
      ELSE
         S := TRUE;
      IF NOT S THEN
         Error(lce_MemHFail,Limit+Delta)
      ELSE
      BEGIN
         { is it being inserted somewhere in the middle of the collection? }
         IF Index <= (Count-1) THEN
         BEGIN
            { shift the existing ptrs up one notch }
            SPtr := MemPtr(Index);
            TPtr := MemPtr(Index+1);
            hMove(SPtr,TPtr,(Count-Index)*SIZEOF(POINTER));
         END;
         { move Item into the 'Index' position }
         TPtr := MemPtr(Index);
         MOVE(Item,TPtr^,SIZEOF(POINTER));
         { add one to the count }
         INC(Count);
      END
   END

END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.AtPut(Index: LONGINT; Item: Pointer);
VAR
   TPtr : POINTER;
BEGIN
   { empty list? }
   IF Count = 0 THEN
      Error(lce_Empty,Index)
   ELSE
   BEGIN
      { check index against list range? }
      IF Index > Count THEN
         Error(lce_IdxGtCount,Index)
      ELSE
      BEGIN
         { valid index? }
         IF ValidIdx(Index) THEN
         BEGIN
            { overwrite current idx pointer with new item }
            TPtr := MemPtr(Index);
            MOVE(Item,TPtr^,SIZEOF(POINTER));
         END
      END
   END
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.Delete(Item: Pointer);
BEGIN
   AtDelete(IndexOf(Item));
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.DeleteAll;
BEGIN
   Count := 0;
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

{ Description: IF you override this method, THEN don't forget to
               call this ancestor method as the first item in the
               new object's own Error method. }

PROCEDURE TLCollection.Error(Code, Info: LONGINT);
BEGIN
   Status := Code;
   ErrorInfo := Info;
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

FUNCTION TLCollection.FirstThat(Test: Pointer): Pointer;
VAR
   Item : POINTER;
   Found : BOOLEAN;
   C : LONGINT;
BEGIN
   Found := FALSE;
   C := 0;
   WHILE (NOT Found) AND (C < Count) DO
   BEGIN
      Item := At(C);
      ASM
         { push the item address }
         push Item.Word[2]
         push Item.Word[0]
         {$IFDEF Windows }
            mov ax, [bp]
            and al, 0FEH
            push ax
         {$ELSE}
            push word ptr [bp]
         {$ENDIF}
         call Test
         mov Found,al
      END;
      INC(C);
   END;
   IF Found THEN
      FirstThat := Item
   ELSE
      FirstThat := NIL
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.ForEach(Action: Pointer);
VAR
   Item : POINTER;
   Idx : LONGINT;
BEGIN
   Idx := 0;
   WHILE Idx < Count DO
   BEGIN
      Item := At(Idx);
      ASM
         { push the item address }
         push word ptr Item.Word[2]
         push word ptr Item.Word[0]
         {$IFDEF Windows }
            mov ax, [bp]
            and al, 0FEH
            push ax
         {$ELSE}
            push word ptr [bp]
         {$ENDIF}
         call Action
      END;
      INC(Idx);
   END;
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.Free(Item: Pointer);
BEGIN
   Delete(Item);
   FreeItem(Item);
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.FreeAll;
VAR
   I : LONGINT;
   APtr : POINTER;
BEGIN
   FOR I := 0 to (Count - 1) DO
   BEGIN
      ErrorInfo := I;
      APtr := At(I);
      IF APtr <> NIL THEN
         FreeItem(At(I));
   END;
   Count := 0;
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.FreeItem(Item: Pointer);
BEGIN
   IF Item <> NIL THEN
      Dispose(PObject(Item),Done);
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

FUNCTION TLCollection.GetItem(VAR S: TStream): Pointer;
BEGIN
   GetItem := S.Get;
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

FUNCTION TLCollection.IndexOf(Item: Pointer): LONGINT;
VAR
   C : LONGINT;
BEGIN
   C := 0;
   WHILE (Item <> At(C)) AND (C < Count) DO
      INC(C);
   IF C >= Count THEN
      IndexOf := -1
   ELSE
      IndexOf := C
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.Insert(Item: Pointer);
BEGIN
   AtInsert(Count,Item);
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

FUNCTION TLCollection.LastThat(Test: Pointer): Pointer;
VAR
   Item : POINTER;
   Found : BOOLEAN;
   C : LONGINT;
BEGIN
   Found := FALSE;
   C := Count-1;
   WHILE (NOT Found) AND (C >= 0) DO
   BEGIN
      Item := At(C);
      ASM
         { push the item address }
         push Item.Word[2]
         push Item.Word[0]
         {$IFDEF Windows }
            mov ax, [bp]
            and al, 0FEH
            push ax
         {$ELSE}
            push word ptr [bp]
         {$ENDIF}
         call Test
         mov Found,al
      END;
      DEC(C);
   END;
   IF Found THEN
      LastThat := Item
   ELSE
      LastThat := NIL
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.Pack;
VAR
   Item : POINTER;
   C : LONGINT;
BEGIN
   C := Count-1;
   WHILE C >= 0 DO
   BEGIN
      Item := At(C);
      IF Item = NIL THEN
         AtDelete(C);
      DEC(C);
   END;
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.PutItem(VAR S: TStream; Item: Pointer);
BEGIN
   S.Put(Item);
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

FUNCTION TLCollection.SetLimit(ALimit: LONGINT) : BOOLEAN;
VAR
   NewMemH : THandle;
   NewAddress : POINTER;
   S : BOOLEAN;
BEGIN
   S := TRUE;
   IF ALimit <> Limit THEN
   BEGIN
      IF ALimit = 0 THEN
      BEGIN
         GlobalUnlock(MemH);
         NewMemH := GlobalFree(MemH);
         NewAddress := NIL;
      END
      ELSE
      BEGIN
         IF MemH = 0 THEN
     	    NewMemH := GlobalAlloc(GMEM_MOVEABLE or GMEM_ZEROINIT,
                                   SIZEOF(POINTER) * ALimit)

         ELSE
         BEGIN
            GlobalUnlock(MemH);
            NewMemH := GlobalReAlloc(MemH,
                                     SIZEOF(POINTER) * ALimit,
                                     GMEM_MOVEABLE or GMEM_ZEROINIT)
         END;
	 IF NewMemH = 0 THEN
            S := FALSE
         ELSE
   	    NewAddress := GlobalLock(NewMemH);
      END;
      IF S THEN
      BEGIN
         MemH := NewMemH;
         Address := NewAddress;
         Limit := ALimit;
      END
   END;
   SetLimit := S;
END;

{ ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLCollection.Store(VAR S: TStream);

   { ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }

   PROCEDURE DoPutItem(P: Pointer); FAR;
   BEGIN
     PutItem(S,P);
   END;

   { ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }

BEGIN  { Store }

   S.Write(Count,SizeOf(Count));
   S.Write(Limit,SizeOf(Limit));
   S.Write(Delta,SizeOf(Delta));
   ForEach(@DoPutItem);

END;   { Store }

{ ::::::::::::::: Private methods :::::::::::::::::::::::::::::::::::::::: }

FUNCTION TLCollection.MemPtr(Idx: LONGINT) : POINTER;
BEGIN
   MemPtr := LIncPtr(Address,Idx * SIZEOF(POINTER))
END;

{ -------------------------------------------------------------------- }

FUNCTION TLCollection.ValidIdx(Idx: LONGINT) : BOOLEAN;
BEGIN
   Status := lce_OK;
   IF Idx < 0 THEN
      Error(lce_IdxLtZero,Idx)
   ELSE
   BEGIN
      IF Idx > Count THEN
         Error(lce_IdxGtCount,Idx)
   END;
   ValidIdx := (Status = lce_OK);
END;

{ :::::::::::::::::: TLSortedCollection methods ::::::::::::::::::::::::: }

CONSTRUCTOR TLSortedCollection.Init(ALimit, ADelta: LONGINT);
BEGIN
   INHERITED Init(ALimit,ADelta);
   Duplicates := False;
END;

{ ----------------------------------------------------------------------- }

CONSTRUCTOR TLSortedCollection.Load(VAR S: TStream);
BEGIN
   INHERITED Load(S);
   S.Read(Duplicates,SizeOf(Boolean));
END;

{ ----------------------------------------------------------------------- }

FUNCTION TLSortedCollection.Compare(Key1, Key2: Pointer): INTEGER;
BEGIN
   Abstract;
END;

{ ----------------------------------------------------------------------- }

FUNCTION TLSortedCollection.IndexOf(Item: Pointer): LONGINT;
var
  I: LONGINT;
BEGIN
   IndexOf := -1;
   IF Search(KeyOf(Item),I) THEN
   BEGIN
      IF Duplicates THEN
         WHILE (I < Count) and (Item <> At(I)) DO
            Inc(I);
      IF I < Count THEN
         IndexOf := I;
   END;
END;

{ ----------------------------------------------------------------------- }

PROCEDURE TLSortedCollection.Insert(Item: Pointer);
var
   I: LONGINT;
BEGIN
   IF NOT Search(KeyOf(Item),I) OR Duplicates THEN
      AtInsert(I,Item);
   ErrorInfo := I;
END;

{ ----------------------------------------------------------------------- }

FUNCTION TLSortedCollection.KeyOf(Item: Pointer): Pointer;
BEGIN
   KeyOf := Item;
END;

{ ----------------------------------------------------------------------- }

FUNCTION TLSortedCollection.Search(Key: Pointer; VAR Index: LONGINT): Boolean;
VAR
  L,
  H,
  I : LONGINT;
  C : INTEGER;
BEGIN
   Search := False;
   L := 0;
   H := Count - 1;
   WHILE L <= H DO
   BEGIN
      I := (L + H) shr 1;
      ErrorInfo := I;
      C := Compare(KeyOf(At(I)), Key);
      IF C < 0 THEN
         L := I + 1
      ELSE
      BEGIN
         H := I - 1;
         IF C = 0 THEN
         BEGIN
           Search := True;
           IF NOT Duplicates THEN
              L := I;
         END;
      END;
   END;
   Index := L;
END;

{ ----------------------------------------------------------------------- }

PROCEDURE TLSortedCollection.Store(VAR S: TStream);
BEGIN
   INHERITED Store(S);
   S.Write(Duplicates, SizeOf(Boolean));
END;

{ :::::::::::::::::: TLStrCollection methods :::::::::::::::::::::::::::: }

FUNCTION TLStrCollection.Compare(Key1, Key2: Pointer): Integer;
BEGIN
   Compare := StrComp(Key1, Key2);
END;

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLStrCollection.FreeItem(Item: Pointer);
BEGIN
   StrDispose(Item);
END;

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

FUNCTION TLStrCollection.GetItem(VAR S: TStream): Pointer;
BEGIN
   GetItem := S.StrRead;
END;

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

PROCEDURE TLStrCollection.PutItem(VAR S: TStream; Item: Pointer);
BEGIN
   S.StrWrite(Item);
END;

{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

BEGIN
END.